//
// "$Id$"
//
// FLUID undo support for the Fast Light Tool Kit (FLTK).
//
// Copyright 1998-2005 by Bill Spitzak and others.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//
// Please report all bugs and problems on the following page:
//
//     http://www.fltk.org/str.php
//

//#include <fltk/run.h>
#include <fltk/Preferences.h>
#include <fltk/string.h>
#include <fltk/MenuBar.h>
#include <fltk/Item.h>

#include "PrefsData.h"
#include "FluidType.h"
#include "undo.h"
#include <sys/stat.h> // mkdir()

#if defined(WIN32) && !defined(__CYGWIN__)
#  include <io.h>
#  include <windows.h>
#  define getpid (int)GetCurrentProcessId
// Visual C++ 2005 incorrectly displays a warning about the use of POSIX APIs
// on Windows, which is supposed to be POSIX compliant...
#  define unlink _unlink
#else
#  include <unistd.h>
#endif // WIN32 && !__CYGWIN__


extern fltk::Item*  undo_item[2];
const int UNDO_ITEM=0;
const int REDO_ITEM=1;

//
// This file implements an undo system using temporary files; ideally
// we'd like to do this in memory, however the current data structures
// and design aren't well-suited...  Instead, we save and restore
// checkpoint files.
//
int Undo::current = 0;
int Undo::last = 0;
int Undo::max = 0;
int Undo::save = -1;			// Last undo level that was saved
char Undo::path[1024]="";
bool Undo::paused=false;
static int current_id = 0;
#define NUM_UNDO_BUFFERS 4

int Undo::undo_index()
{
  return current_id;
}

static bool init_undo_dirs()
{
  // get the data path or just return.
  if (!Undo::path[0]) prefs.getUserdataPath(Undo::path, sizeof(Undo::path));
  if(!Undo::path[0])
    return 0;
  
  char buf [256];
  sprintf(buf, "%sundo", Undo::path);
  int slen = strlen(buf);
  char* s;
  
  // create each directory .../fluid2/undo<N>
  for(int i = 1; i <= NUM_UNDO_BUFFERS; i++)
  {
    s = buf + slen;
    sprintf(s, "%d", i);
    mkdir(buf, 0777);
  }
  return true;
}

static int undo_exists(int i)
{
  // save position in path scratch pad
  FILE* fp;
  char* posn;
  char buf[256];
  strcpy(buf, Undo::path);
  posn = buf + strlen(buf);
  sprintf(posn, "undo%d/%d.fl", current_id, i);
  fp = fopen(buf, "r");
  if(fp) fclose(fp);
  return fp != 0 ? true : false;
}

// called by fluid.cxx main() to create and/or clean out 
// a set of undo folders in a rolling array indexed by
// the global undo_id file in the undo folder.
void init_undo()
{
  char buf[256];
  int next_id;
  FILE * fp;

#warning
  if(!init_undo_dirs()) 
    // init prefs here or warn that fluid will need to be
    // restarted for undos to work?
    return;
  
  
  // try to get the current id
  // This creates a new id from a file named "undo_id" in the
  // .fltk/fltk.org/fluid2 directory where the undos are.
    
  strcpy(buf, Undo::path);
  strcat(buf, "undo_id");
  fp = fopen(buf, "r");
  
  if(fp) // it exists
  {  
    // get the current index/id for the undo folder to write into
    fscanf(fp, "%d", &current_id);
    fclose(fp);
  }
  else // it doesn't exist
  {
    current_id = 1;
  }    
  // modulus NUM_UNDO_BUFFERS + 1
  if((current_id == 0) || (current_id > NUM_UNDO_BUFFERS))
    current_id = 1;
  
  // set up the id for the next time we open fluid
  next_id = current_id + 1;
  fp = fopen(buf, "w"); 
  if(! fp) return;
  fprintf(fp, "%d", next_id);
  fclose(fp);
  
  sprintf(buf, Undo::path);
  
  for(int i = 0; i < 1000; i++)
  {
    if(undo_exists(i))
      remove(Undo::filename(current_id, i, buf, 256));
    else
      break;
  }
}


// Return the undo filename
const char *Undo::filename(int index, int level, char *buf, int bufsize) {

  if (!path[0]) prefs.getUserdataPath(path, sizeof(path));
  snprintf(buf, bufsize, "%sundo%d/%d.fl", path, index, level);
  return buf;
}

void Undo::update_saved() {
  if (modflag && current <= save) save = -1;
  else if (!modflag) save = current;
}

// Redo menu callback
void Undo::redo_cb(fltk::Widget *, void *) {
  char	fname[1024];			// Undo checkpoint file
  if (current >= last) return;

  suspend();
  if (!read_file(filename(undo_index(), current + 1, fname, sizeof(fname)), 0)) {
    // Unable to read checkpoint file, don't redo...
    resume();
    return;
  }

  current ++;

  // Update modified flag...
  modflag = current != save;

  // Update undo/redo menu items...
  if (current >= last) undo_item[REDO_ITEM]->deactivate();
  undo_item[UNDO_ITEM]->activate();
}

// Undo menu callback
void Undo::undo_cb(fltk::Widget *, void *) {
  char	fname[1024];			// Undo checkpoint file
  
  if (current <= 0) return;

  if (current == last) {
    write_file(filename(undo_index(), current, fname, sizeof(fname)));
  }

  suspend();
  if (!read_file(filename(undo_index(), current - 1, fname, sizeof(fname)), 0)) {
    // Unable to read checkpoint file, don't undo...
    resume();
    return;
  }

  current--;

  // Update modified flag...
  modflag = (current != save);

  // Update undo/redo menu items...
  if (current <= 0) undo_item[UNDO_ITEM]->deactivate();
  undo_item[REDO_ITEM]->activate();
  resume();
}

// Save current file to undo buffer
void Undo::checkpoint() {
  char	fname[1024];			// Undo checkpoint filename

//  printf("undo_checkpoint(): current=%d, undo_paused=%d, modflag=%d\n",
//         current, undo_paused, modflag);

  // Don't checkpoint if undo_suspend() has been called...
  if (paused) return;

  // Save the current UI to a checkpoint file...
  if (!write_file(filename (undo_index(), current, fname, sizeof(fname)))) 
  {
    // Don't attempt to do undo stuff if we can't write a checkpoint file...
    if (!write_file(filename (undo_index(), current, fname, sizeof(fname))))
      perror(fname);
    return;
  }

  // Update the saved level...
  if (modflag && current <= save) save = -1;
  else if (!modflag) save = current;

  // Update the current undo level...
  current ++;
  last = current;
  if (current > max) max = current;

  // Enable the Undo and disable the Redo menu items...
  undo_item[UNDO_ITEM]->activate();
  undo_item[REDO_ITEM]->deactivate();
}

// Clear undo buffer
void Undo::clear() {
  char	fname[1024];			// Undo checkpoint filename

  // Remove old checkpoint files...
  for (int i = 0; i <= max; i ++) {
    unlink(filename (undo_index(), i, fname, sizeof(fname)));
  }

  // Reset current, last, and save indices...
  current = last = max = 0;
  if (modflag) save = -1;
  else save = 0;
}

// remove the last set checkpoint
void Undo::remove_last() {
    current --;    
    last --;
    if (current <=0)   
	undo_item[UNDO_ITEM]->deactivate();

}

//
// End of "$Id$".
//
